import textwrap as tw
import pytest
from zgulde.flashcards.__main__ import make_cards, FlashCard

def test_card_parsing():
    content = tw.dedent('''
    here is the front
    ---
    and this is the back
    ===
    this is another card
    ---
    this is the second back
    ''').strip()

    cards = make_cards(content)

    assert len(cards) == 2

    first, second = cards

    assert first.front == 'here is the front'
    assert first.back == 'and this is the back'

    assert second.front == 'this is another card'
    assert second.back == 'this is the second back'

def test_whitespace_doesnt_matter_in_card_parsing():
    content = tw.dedent('''

    here is the front

    ---

    and this is the back

    ===

    this is another card


    ---

    this is the second back
    ''').strip()

    cards = make_cards(content)

    assert len(cards) == 2

    first, second = cards

    assert first.front == 'here is the front'
    assert first.back == 'and this is the back'

    assert second.front == 'this is another card'
    assert second.back == 'this is the second back'



#!/usr/bin/env python3

import fileinput

from typing import List
from pprint import pprint

import os
import sys
import textwrap

import itertools as it

FILE_EXTENSIONS = dict(
    sh=dict(comment='#', language='bash'),
    py=dict(comment='#', language='python'),
    js=dict(comment='//', language='javascript'),
    sql=dict(comment='--', language='sql'),
    php=dict(comment='//', language='php'),
)

def prev_and_next(xs):
    prevs, items, nexts = it.tee(xs, 3)
    return zip(it.chain([None], prevs),
               items,
               it.chain(it.islice(nexts, 1, None), [None]))

def make_categorizor(doc_marker):
    def is_code(line):
        return line is not None and not line.startswith(doc_marker) and line != ''

    def is_docs(line):
        return line is not None and line.startswith(doc_marker)

    def categorize(x):
        prev, line, next = x
        if is_docs(line):
            return 'docs'
        elif is_code(line):
            return 'code'
        elif line == '' and is_code(prev) and is_code(next):
            return 'code'
        elif line == '' and is_docs(prev) and is_docs(next):
            return 'docs'
        else:
            return 'ignore'
    return categorize

def main(fp, language, doc_marker):
    lines = map(lambda line: line.rstrip(), fileinput.input(fp))
    keyfn = make_categorizor(doc_marker)
    chars_to_skip = len(doc_marker)

    for section_type, contents in it.groupby(prev_and_next(lines), keyfn):
        if section_type == 'ignore':
            pass
        elif section_type == 'code':
            print(f'```{language}')
            for _, line, _ in contents:
                print(line)
            print('```')
            print()
        elif section_type == 'docs':
            for _, line, _ in contents:
                print(line[chars_to_skip:].lstrip())
            print()
        else:
            raise Error(f'Unknown section type: {section_type}')

def get_defaults(args):
    fp = args.file if args.file is not None else '-'
    extension = fp.split('.')[-1]
    language = args.language
    doc_marker = args.doc_marker

    if language is None and extension in FILE_EXTENSIONS.keys():
        language = FILE_EXTENSIONS[extension]['language']
    elif language is None:
        language = ''

    if doc_marker is None and extension in FILE_EXTENSIONS.keys():
        doc_marker = FILE_EXTENSIONS[extension]['comment'] + ':'
    elif doc_marker is None:
        doc_marker = '#:'

    return fp, language, doc_marker

def show_defaults():
    print('ext marker language')
    for extension, d in FILE_EXTENSIONS.items():
        print('{:>3} {:^6} {}'.format(extension, d['comment'] + ':', d['language']))


if __name__ == '__main__':
    import argparse
    import textwrap

    description = '''
    Transform source code -> markdown.

    Specially marked lines (those that start with the doc_marker) are treated as
    markdown, and everything else is turned into a fenced code block.
    '''

    parser = argparse.ArgumentParser()
    parser.prog = 'python -m zgulde.em'
    parser.description = textwrap.dedent(description)
    parser.add_argument('file', metavar='FILE', help='filepath', nargs='?')
    parser.add_argument('-m', '--doc-marker', help='indicator for a documentation line')
    parser.add_argument('-l', '--language', help='language for fenced code blocks')
    parser.add_argument('--show-defaults', action='store_true', help='show defaults for each file extension and exit')
    args = parser.parse_args()

    if args.show_defaults:
        show_defaults()
        sys.exit(0)

    fp, language, doc_marker = get_defaults(args)
    main(fp, language, doc_marker)



from typing import Callable, List, Iterable, Dict, Any, TypeVar, Tuple, Optional, Sequence
import itertools as it
from functools import reduce, partial
import collections
import hashlib

def md5(s):
    return hashlib.md5(s.encode('utf8')).hexdigest()
def sha256(s):
    return hashlib.sha256(s.encode('utf8')).hexdigest()

def slurp(fp):
    with open(fp) as f:
        return f.read()

def spit(fp, content, mode='w+'):
    with open(fp) as f:
        f.write(content)

def pluck(d: Dict, *ks: str):
    '''
    Pluck specified values from a dictionary.

    This can make for concise code together with unpacking syntax.

    Example
    -------

    >>> d = {'x': 1, 'y': 2}
    >>> pluck(d, 'x')
    [1]
    >>> pluck(d, 'y', 'x')
    [2, 1]
    >>> x, y = pluck(d, 'x', 'y')
    >>> x
    1
    >>> y
    2
    '''
    return [d[k] for k in ks]

def comp(*fns: Callable):
    '''
    Returns a function that is the passed functions composed together. Functions
    are applied from right to left.

    comp(f, g)(x) == f(g(x))
    comp(f, g, h)(x) == f(g(h(x)))

    Example
    -------

    >>> def double(n):
    ...     print(f'Doubling {n}...')
    ...     return n * 2
    >>> def inc(n):
    ...     print(f'Incrementing {n}...')
    ...     return n + 1
    >>> double_plus_one = comp(double, inc)
    >>> double_plus_one(3)
    Incrementing 3...
    Doubling 4...
    8
    >>> double_plus_one(5)
    Incrementing 5...
    Doubling 6...
    12
    '''
    return lambda x: reduce(lambda x, f: f(x), reversed(fns), x)

def partition(xs: Sequence, chunksize: int):
    '''
    Partition a sequence into smaller subsequences

    Returns a generator that yields the smaller partitions

    >>> letters = 'abcdefghijklm'
    >>> partition(letters, 2)
    <generator object partition at ...>
    >>> list(partition(letters, 2))
    ['ab', 'cd', 'ef', 'gh', 'ij', 'kl', 'm']
    >>> list(partition(letters, 3))
    ['abc', 'def', 'ghi', 'jkl', 'm']
    '''
    for i in range(0, len(xs), chunksize):
        yield xs[i:i + chunksize]

chunk = partition

def pipe(v: Any, *fns: Callable):
    '''
    Thread a value through one or more functions.

    Functions are applied left to right

    >>> def double(n):
    ...     print(f'Doubling {n}...')
    ...     return n * 2
    >>> def inc(n):
    ...     print(f'Incrementing {n}...')
    ...     return n + 1
    >>> pipe(3, inc, double)
    Incrementing 3...
    Doubling 4...
    8
    '''

    return reduce(lambda x, f: f(x), fns, v)

def take(xs, n):
    return list(it.islice(xs, n))

def tail(xs, n):
    return list(collections.deque(xs, maxlen=n))

def drop(xs, n):
    return it.islice(xs, n, None)

def prepend(xs, v):
    return it.chain([v], xs)

def append(xs, v):
    return it.chain(xs, [v])

A = TypeVar('A')
def and_prev(xs: Iterable[A]) -> Iterable[Tuple[Optional[A], A]]:
    '''
    Return each item of the iterable xs, along with the previous item.

    The first previous item is None.

    Returns
    -------

    An iterable of tuples, where each tuple is the previous item and the current
    item.

    >>> list(and_prev([1, 2, 3]))
    [(None, 1), (1, 2), (2, 3)]
    '''
    return zip(prepend(xs, None), xs)

def and_next(xs: Iterable[A]) -> Iterable[Tuple[A, Optional[A]]]:
    '''
    Return each of the items in the iterable xs, along with the next item.

    When the iterable is exhausted, the last item is None.

    Returns
    -------

    An iterable of tuples, where each tuple is the current item and the next
    item.

    >>> list(and_next([1, 2, 3]))
    [(1, 2), (2, 3), (3, None)]
    '''
    return it.zip_longest(xs, drop(xs, 1))

def prev_and_next(xs: Iterable[A]) -> Iterable[Tuple[Optional[A], A, Optional[A]]]:
    '''
    Return each of the items in the iterable xs, along with the previous and
    next elements.

    For the first item, the previous is None, for the last item, the next is
    None.

    Returns
    -------

    An iterable of tuples, where each tuple is (prev, current, next)

    >>> list(prev_and_next([1, 2, 3]))
    [(None, 1, 2), (1, 2, 3), (2, 3, None)]
    '''
    prev, current, next = it.tee(xs, 3)
    return zip(prepend(prev, None), current, drop(append(next, None), 1))



from IPython.display import display, HTML as nb_HTML
from prompt_toolkit import print_formatted_text, HTML as cli_HTML
import re


def hl_matches(
    regexp: str,
    subject: str,
    start: str,
    end: str,
    groupstart: str,
    groupend: str,
    initial_output="",
    closing_output="",
) -> str:
    """
    Does not handle nested groups.
    """

    m = re.search(regexp, subject)

    if m is None:
        print("No matches!")
        return subject

    n_matches = len(m.groups())
    overall_start, overall_end = m.span()

    spans = [m.span(n) for n in range(1, n_matches + 1)]
    starts = [span[0] for span in spans]
    ends = [span[1] for span in spans]

    output = initial_output

    for i, c in enumerate(subject):
        if i == overall_start:
            output += start

        if i in starts:
            output += groupstart

        if i in ends:
            output += groupend

        if i == overall_end:
            output += end

        output += c

    if len(subject) in ends:
        output += groupend
    if len(subject) == overall_end:
        output += end

    output += closing_output
    return output


def hl_matches_cli(regexp, subject):
    """
    Highlight a regular expressions matches in a string.

    Currently only works in terminal environments (read: this function won't
    work in a jupyter notebook) and does not handle nested groups.

    Uses python's `re.search` under the hood.

    >>> from zgulde.hl_matches import hl_matches
    >>> hl_matches(r'^.', 'abc')
    >>> hl_matches(r'^(\d).*?(\d+)$', '123 broadway st san antonio tx 78205')

    The entire match will be highlighted in bold, the capture groups will be
    red.
    """
    output = hl_matches(
        regexp,
        subject,
        start="<b>",
        end="</b>",
        groupstart="<firebrick>",
        groupend="</firebrick>",
    )
    print_formatted_text(cli_HTML(output))


def hl_matches_nb(regexp, subject):
    """
    Highlight a regular expressions matches in a string. Does not handle nested
    groups.

    Uses python's `re.search` under the hood.

    >>> from zgulde.hl_matches import hl_matches
    >>> hl_matches(r'^.', 'abc')
    >>> hl_matches(r'^(\d).*?(\d+)$', '123 broadway st san antonio tx 78205')

    The entire match will be underlined, and the capture groups will be red.
    """
    output = hl_matches(
        regexp,
        subject,
        start='<span style="text-decoration: underline">',
        end="</span>",
        groupstart='<span style="color: firebrick">',
        groupend="</span>",
        initial_output='<div style="font-family: monospace; letter-spacing: 3px; font-size: 24px; line-height: 36px;">',
        closing_output="</div>",
    )
    return nb_HTML(output)



import pandas as pd
from pydataset import data

from .plotting import *
from .modeling import *

# mode aggregation
# tips.groupby('size').agg(stats.mode).apply(lambda s: s.apply(lambda v: v[0][0]))

# better interaction w/ sklearn.metrics.confusion_matrix, e.g. auto labelling
def better_confusion_matrix(actual, predicted, labels=None):
    cm = pd.crosstab(actual, predicted)
    if labels is not None:
        cm.index = pd.Index(labels, name='Actual')
        cm.columns = pd.Index(labels, name='Predictions')
    return cm


def labelled_confusion_matrix():
    matrix = [['TP',                         'FN (Type II Error)', 'Recall = TP / (TP + FN)'],
              ['FP (Type I Error)',          'TN',                 ''],
              ['Precision = TP / (TP + FP)', '',                   '']]
    return pd.DataFrame(matrix,
                        columns=['Predicted +', 'Predicted -', ''],
                        index=['Actual +', 'Actual -', ''])

def rdatasets():
    datasets = data()
    shapes = datasets.apply(lambda row: data(row.dataset_id).shape, axis=1)\
                     .apply(pd.Series)\
                     .rename({0: 'nrows', 1: 'ncols'}, axis=1)
    return pd.concat([datasets, shapes], axis=1)

class MyRange:
    """
    A simple class that defines a convenient (if not conventional) range syntax

    >>> r = MyRange()
    >>> r[1:10]
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    >>> r[1:10:2]
    [1, 3, 5, 7, 9]
    """
    def __getitem__(_, aslice: slice):
        start = aslice.start if aslice.start is not None else 0
        step = aslice.step if aslice.step is not None else 1
        stop = aslice.stop + 1 if aslice.stop is not None else 1
        return list(range(start, stop, step))

# https://stackoverflow.com/questions/50559078/generating-random-dates-within-a-given-range-in-pandas
def rand_dates(start, end, n):
    start_u = pd.to_datetime(start).value // 10**9
    end_u = pd.to_datetime(end).value // 10**9
    return pd.DatetimeIndex((10**9*np.random.randint(start_u, end_u, n)).view('M8[ns]'))

import graphviz
from sklearn.tree import export_graphviz, DecisionTreeClassifier

def viz_dtree(X, y, **kwargs):
    tree = DecisionTreeClassifier(**kwargs).fit(X, y)
    feature_names = X.columns
    class_names = sorted(y.unique())
    dot = export_graphviz(
        tree,
        out_file=None,
        feature_names=feature_names,
        class_names=class_names, # target value names
        special_characters=True,
        filled=True,             # fill nodes w/ informative colors
        impurity=False,          # show impurity at each node
        leaves_parallel=True,    # all leaves at the bottom
        proportion=True,         # show percentages instead of numbers at each leaf
        rotate=True,             # left to right instead of top-bottom
        rounded=True,            # rounded boxes and sans-serif font
    )
    graph = graphviz.Source(dot, filename='dtree', format='png')
    graph.view(cleanup=True)




'''
A module that provides all the common data science imports.

For example:

>>> from zgulde.ds_imports import *

and all the common data science imports will be in your namespace.

Intended for quick iteration / experimentation, and explicit imports should be
used in production code.
'''

import sys
import os
import itertools as it
from functools import reduce, partial
import math
from math import sqrt, factorial

import pandas as pd
from pandas import DataFrame, Series
import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import seaborn as sns

from scipy import stats
from scipy.stats import binom, norm, geom, poisson

import sklearn
from sklearn.model_selection import GridSearchCV, train_test_split, cross_val_score
from sklearn.linear_model import LinearRegression, LogisticRegression, Ridge
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.svm import SVC, SVR
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.cluster import DBSCAN, KMeans, dbscan, k_means
from sklearn.preprocessing import scale, StandardScaler
from sklearn.metrics import r2_score, mean_absolute_error, mean_squared_error
from sklearn.metrics import accuracy_score, precision_score, recall_score

from sklearn.pipeline import Pipeline

from pydataset import data

from zgulde import partition, chunk, comp, pluck, extend_pandas
from zgulde.ds_util import *

# check if we're running interactively
if hasattr(sys, 'ps1'):
    # turn on interactive mode in matplotlib
    plt.ion()

r = MyRange()
tips = data('tips')
mpg = data('mpg')
mtcars = data('mtcars')
swiss = data('swiss')
iris = data('iris').cleanup_column_names()

# plotting style defaults
plt.rc('patch', edgecolor='black', force_edgecolor=True, facecolor='firebrick')
plt.rc('axes', grid=True)
plt.rc('grid', linestyle=':', linewidth=.8, alpha=.7)
plt.rc('axes.spines', right=False, top=False)
plt.rc('figure', figsize=(11, 8))
plt.rc('font', size=12.0)
plt.rc('hist', bins=25)



from .__main__ import make_categorizor, prev_and_next
import textwrap
import os
import pytest
from subprocess import check_output

@pytest.fixture
def extract_markdown():
    def _extract_markdown(input, expected, args=None):
        cmd = 'python -m zgulde.em'
        input = textwrap.dedent(input).strip()
        expected = textwrap.dedent(expected).strip()
        if args is not None:
            cmd += ' ' + args
        output = check_output(cmd.split(),
                              input=input.encode()).decode().strip()
        print('----- actual')
        print(output)
        print('----- expected')
        print(expected)
        return output, expected
    yield _extract_markdown

def test_categorizor():
    lines = textwrap.dedent('''
    #: # Hello, World
    #:
    #: This is a demo.

    foo

    bar

    baz
    ''').strip().split('\n')
    categorize = make_categorizor('#:')
    categories = [categorize(line) for line in prev_and_next(lines)]
    assert categories == ['docs'] * 3 + ['ignore'] + ['code'] * 5

def test_docs_and_code(extract_markdown):
    actual, expected = extract_markdown('''
        #: docs
        #: docs
        #: docs
        code
        code
        ''',
        '''
        docs
        docs
        docs

        ```
        code
        code
        ```
        '''
    )
    assert actual == expected

def test_just_docs(extract_markdown):
    actual, expected = extract_markdown('''
        #: docs
        ''',
        '''
        docs
        '''
    )
    assert actual == expected


def test_just_code(extract_markdown):
    actual, expected = extract_markdown('''
        code
        ''',
        '''
        ```
        code
        ```
        '''
    )
    assert actual == expected

def test_blank_lines_are_preserved_within_docs_or_code(extract_markdown):
    actual, expected = extract_markdown('''
        #: docs
        #:
        #: docs

        code

        code
        ''',
        '''
        docs

        docs

        ```
        code

        code
        ```
        '''
    )
    assert actual == expected

def test_blank_lines_between_docs_and_code_are_auto_inserted_if_not_present(extract_markdown):
    actual, expected = extract_markdown('''
        #: docs
        code
        #: docs
        code
        ''',
        '''
        docs

        ```
        code
        ```

        docs

        ```
        code
        ```
        '''
    )
    assert actual == expected

def test_multiple_blank_lines_between_docs_and_code_are_ignored(extract_markdown):
    actual, expected = extract_markdown('''
        #: docs


        code


        #: docs


        code
        ''',
        '''
        docs

        ```
        code
        ```

        docs

        ```
        code
        ```
        '''
    )
    assert actual == expected

def test_2_or_more_blanks_in_code_create_seperate_blocks(extract_markdown):
    actual, expected = extract_markdown('''
        code


        code
        ''',
        '''
        ```
        code
        ```

        ```
        code
        ```
        '''
    )
    assert actual == expected

def test_it_highlights_adds_the_given_language_to_fenced_code_blocks(extract_markdown):
    actual, expected = extract_markdown('''
        code
        ''',
        '''
        ```bash
        code
        ```
        ''',
        args='-l bash'
    )
    assert actual == expected
    actual, expected = extract_markdown('''
        code
        ''',
        '''
        ```bash
        code
        ```
        ''',
        args='--language bash'
    )
    assert actual == expected

def test_it_accepts_a_different_doc_marker(extract_markdown):
    actual, expected = extract_markdown('''
        ## Here is some documentation
        this is code
        ''',
        '''
        Here is some documentation

        ```
        this is code
        ```
        ''',
        args='-m ##'
    )
    assert actual == expected
    actual, expected = extract_markdown('''
        ## Here is some documentation
        this is code
        ''',
        '''
        Here is some documentation

        ```
        this is code
        ```
        ''',
        args='--doc-marker ##'
    )
    assert actual == expected




import math
from typing import Callable

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# TODO: refactor some of these to iterate over .groupby instead of unique values


def plot_dual_axis(df: pd.DataFrame, x: str) -> Callable:
    """
    plot_dual_axis(df, 'x')('y1')('y2')
    """
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()

    def plot_y2(y, *args, **kwargs):
        ax2.plot(df[x], df[y], *args, **kwargs)
        return fig

    def plot_y1(y, *args, **kwargs):
        ax1.plot(df[x], df[y], *args, **kwargs)
        return plot_y2

    return plot_y1


def plot_3d(df, x, y, z, g):
    fig = plt.figure()
    ax = Axes3D(fig)
    for v in sorted(df[g].unique()):
        subset = df[df[g] == v]
        ax.scatter(subset[x], subset[y], subset[z], label=v)
    ax.set(xlabel=x, ylabel=y, zlabel=z)
    ax.legend(title=g)


def plot_dual_axis(df: pd.DataFrame, x: str):
    "plot_dual_axis(df, 'x')('y1', marker='x')('y2', color='firebrick')"
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()

    def plot_y2(y, *args, **kwargs):
        ax2.plot(df[x], df[y], *args, **kwargs)
        return fig

    def plot_y1(y, *args, **kwargs):
        ax1.plot(df[x], df[y], *args, **kwargs)
        return plot_y2

    return plot_y1


def top_n(s: pd.Series, n=3, other_val="Other"):
    top_n = s.value_counts().index[:n]
    return pd.Series(np.where(s.isin(top_n), s, other_val), name=s.name)


def plot_hist_by_group(x: pd.Series, g: pd.Series, *args, **kwargs):
    """
    >>> mpg = data('mpg')
    >>> hist_by_group(mpg.hwy, mpg.cyl)
    """
    fig, axs = plt.subplots(2, 2) if g.nunique() > 2 else plt.subplots(1, 2)
    fig.suptitle(f"Distribution of {x.name} by {g.name}")
    g = top_n(g, 3)
    g.index = x.index

    for ax, v in zip(axs.ravel(), g.unique()):
        x_g = x[g == v]

        ax.hist(x_g, color="pink", *args, **kwargs)
        ax.set_title(v)

        # mean + ci
        xbar = x_g.mean()
        z = 1.96  # 95% ci
        ci = z * (x_g.std() / math.sqrt(x_g.shape[0]))
        ub, lb = xbar + ci, xbar - ci

        ymin, ymax = ax.get_ylim()
        ax.vlines(xbar, ymin, ymax, ls="--", color="gray")
        ax.vlines([lb, ub], ymin, ymax, ls=":", color="gray")

    return fig, axs


def plot_bar_by_group(x: pd.Series, g: pd.Series, aggfunc="mean", *args, **kwargs):
    """
    >>> mpg = data('mpg')
    >>> plot_bar_by_group(mpg.hwy, mpg['class'])
    """
    g = top_n(g, 3)
    fig, ax = plt.subplots()
    x.groupby(g).agg(aggfunc).plot.bar(ax=ax, color="pink", width=1)
    ax.set(title=f"{aggfunc} of {x.name} by {g.name}")
    if aggfunc == "mean":
        xbar = x.agg(aggfunc)
        z = 2.58  # 99% ci
        ci = z * (x.std() / math.sqrt(x.shape[0]))
        ub, lb = xbar + ci, xbar - ci
        ax.hlines(xbar, -0.5, 3.5, ls="--", color="gray")
        ax.hlines([lb, ub], -0.5, 3.5, ls=":", color="gray")
    else:
        ax.hlines(x.agg(aggfunc), -0.5, 3.5, ls="--", color="gray")
    return fig, ax


def plot_scatter_by_group(df: pd.DataFrame, x: str, y: str, g: str):
    for name, subset in df.groupby(g):
        plt.scatter(subset[x], subset[y], label=name)
    plt.xlabel(x)
    plt.ylabel(y)
    plt.legend(title=g)


# WIP below
arrowstyles = [
    "-",
    "->",
    "-[",
    "|-|",
    "-|>",
    "<-",
    "<->",
    "<|-",
    "<|-|>",
    "fancy",
    "simple",
    "wedge",
]


def annotate(s, xy, xytext, style="->"):
    return plt.annotate(s, xy=xy, xytext=xytext, arrowprops={"arrowstyle": style})


def arrow(fromxy, toxy, style="->"):
    plt.annotate(
        "",
        xy=(11, 0.07),
        xytext=(30, 0.07),
        xycoords="data",
        textcoords="data",
        arrowprops={"arrowstyle": "|-|"},
    )



from .ds_util import MyRange
import pytest

@pytest.fixture()
def r():
    return MyRange()

def test_it_generates_a_range(r):
    assert r[5:8] == [5, 6, 7, 8]
    assert r[5:8:2] == [5, 7]
    assert r[:5] == [0, 1, 2, 3, 4, 5]
    assert r[5:] == []



import math
import numpy as np
import pandas as pd

from typing import List, Tuple, Dict, Union

import sklearn
from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error
from sklearn.metrics import accuracy_score, precision_score, recall_score
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.linear_model import LinearRegression, LogisticRegression, Ridge
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.svm import SVC, SVR
from sklearn.model_selection import GridSearchCV

def residuals(model, X, y):
    return y - model.predict(X)

def tree(df, rformula, **kwargs):
    X, y = df.rformula(rformula)
    if np.issubdtype(y.dtype, np.number):
        return dtree_regressor(X, y, **kwargs)
    else:
        y = y.astype('category').cat.codes
        return dtree_classifier(X, y, **kwargs)

def lr(df, rformula, **kwargs):
    X, y = df.rformula(rformula)
    if np.issubdtype(y.dtype, np.number):
        return linear_regression(X, y, **kwargs)
    else:
        y = y.astype('category').cat.codes
        return logistic_regression(X, y, **kwargs)

def logistic_regression(X, y, **kwargs):
    lr = LogisticRegression(**kwargs).fit(X, y)
    yhat = lr.predict(X)
    coefs = pd.Series(dict(zip(X.columns, lr.coef_[0])))
    return {
        'model': lr,
        'accuracy': accuracy_score(y, yhat),
        'precision': precision_score(y, yhat, average=None),
        'recall': recall_score(y, yhat, average=None),
        'coef': coefs.sort_values()
    }

def dtree_classifier(X, y, **kwargs):
    tree = DecisionTreeClassifier(**kwargs).fit(X, y)
    yhat = tree.predict(X)
    features = pd.Series(dict(zip(X.columns, tree.feature_importances_)))
    return {
        'model': tree,
        'accuracy': accuracy_score(y, yhat),
        'precision': precision_score(y, yhat, average=None),
        'recall': recall_score(y, yhat, average=None),
        'feature_importances': features.sort_values()
    }

def dtree_regressor(X, y, **kwargs):
    tree = DecisionTreeRegressor(**kwargs).fit(X, y)
    yhat = tree.predict(X)
    mse = mean_squared_error(y, yhat)
    features = pd.Series(dict(zip(X.columns, tree.feature_importances_)))
    return {
        'model': tree,
        'r2': r2_score(y, yhat),
        'mse': mse,
        'rmse': math.sqrt(mse),
        'mae': mean_absolute_error(y, yhat),
        'feature_importances': features.sort_values(),
    }

def linear_regression(X, y, **kwargs):
    lr = LinearRegression(**kwargs).fit(X, y)
    yhat = lr.predict(X)
    mse = mean_squared_error(y, yhat)
    coefs = pd.Series(dict(zip(X.columns, lr.coef_)))
    return {
        'model': lr,
        'r2': r2_score(y, yhat),
        'mse': mse,
        'rmse': math.sqrt(mse),
        'mae': mean_absolute_error(y, yhat),
        'coef': coefs.sort_values(),
    }


def cv_results_to_df(grid: GridSearchCV) -> pd.DataFrame:
    results = grid.cv_results_
    params_and_scores = zip(results["params"], results["mean_test_score"])
    df = pd.DataFrame([dict(**p, score=s) for p, s in params_and_scores])
    df["model"] = grid.best_estimator_.__class__.__name__
    return df


GridCandidates = List[
    Tuple[sklearn.base.BaseEstimator, Dict[str, List[Union[str, int]]]]
]


def multi_grid_search(models: GridCandidates, X, y, cv=4) -> pd.DataFrame:
    return pd.concat([
        cv_results_to_df(GridSearchCV(model, params, cv=cv).fit(X, y))
        for model, params in models
    ])

def inspect_coefs(lm, X):
    coef = lm.coef_
    if coef.shape[0] > 1:
        return pd.DataFrame(coef, index=lm.classes_, columns=X.columns)
    else:
        return pd.Series(dict(zip(X.columns, lm.coef_.ravel()))).sort_values()

def inspect_feature_importances(tree, X):
    return pd.Series(dict(zip(X.columns, tree.feature_importances_))).sort_values()


classification_models = [
    (DecisionTreeClassifier(), {"max_depth": range(1, 11)}),
    (KNeighborsClassifier(), {"n_neighbors": range(1, 11)}),
    (LogisticRegression(), {"C": [0.01, 0.1, 1, 10, 100, 1000], "solver": ["lbfgs"]}),
    (SVC(), {"kernel": ["rbf", "linear"]}),
    (SVC(), {"kernel": ["poly"], "degree": [2]}),
]

regression_models = [
    (DecisionTreeRegressor(), {"max_depth": range(1, 11)}),
    (KNeighborsRegressor(), {"n_neighbors": range(1, 11)}),
    (LinearRegression(), {}),
    (Ridge(), {"alpha": [0.01, 0.1, 1, 10, 100, 1000]}),
    (SVR(), {"kernel": ["rbf", "linear"]}),
    (SVR(), {"kernel": ["poly"], "degree": [2]}),
]




from sys import exit
from os import path, system
from argparse import ArgumentParser
from typing import List, NamedTuple
from IPython import embed

FlashCard = NamedTuple('FlashCard', front=str, back=str)
Deck = List[FlashCard]

def die(msg, code=1):
    print(msg)
    exit(1)

def make_cards(content: str) -> Deck:
    card_contents = content.split('\n===\n')
    faces = [card.strip().split('\n---\n') for card in card_contents]
    cards = [FlashCard(front.strip(), back.strip()) for front, back in faces]
    return cards

def review_cards(cards: Deck):
    for card in cards:
        system('clear')
        print(card.front)
        input('Answer? ')
        print('---')
        print()
        print(card.back)
        input('(c)orrect (i)ncorrect? ')

if __name__ == '__main__':
    parser = ArgumentParser(prog='zgulde.flashcards')
    parser.add_argument('flashcard_file')
    args = parser.parse_args()

    print(args.flashcard_file)

    if not path.exists(args.flashcard_file):
        die(f'Error: could not find file {args.flashcard_file}')

    with open(args.flashcard_file) as f:
        content = f.read()

    cards = make_cards(content)
    review_cards(cards)
